from fastapi.exceptions import HTTPException
from motor.motor_asyncio import AsyncIOMotorClient
from pymongo.errors import ServerSelectionTimeoutError


class MongoDB:
    def __init__(self, connection_url: str, db_name: str):
        self.client = None
        self.connection_url = connection_url
        self.db_name = db_name

    async def connect(self):
        try:
            self.client = AsyncIOMotorClient(self.connection_url)
        except ServerSelectionTimeoutError:
            raise HTTPException(
                status_code=500,
                detail="Failed to connect to the database. Check your connection settings.",
            )

    async def close(self):
        if self.client:
            self.client.close()

    def get_database(self):
        if not self.client:
            raise HTTPException(
                status_code=500,
                detail="Database connection not established. Call connect() method first.",
            )
        return self.client[self.db_name]



# -*- coding: UTF-8 -*-
import logging
import logging.handlers
import os
from abc import ABCMeta, abstractmethod
from functools import lru_cache

import dotenv
import watchtower
from beanie import init_beanie
from fastapi import FastAPI, status
from fastapi.exceptions import RequestValidationError
from fastapi.middleware.cors import CORSMiddleware

from . import database
from .const import LOG_FMT, LOG_FMT_NO_DT, LOG_DT_FMT, DEFAULT_LOGGER_NAME
from .exception.base_exception import InternalBaseException
from .ext.amazon import aws
from .http.responses import async_response
from .utils import update_dict_with_cast


class BaseFactory(metaclass=ABCMeta):
    DEFAULT_APP_NAME = ""
    API_VERSION = "v0.0.0"

    @abstractmethod
    def init_modules(self, app):
        """
        Each factory should define what modules it wants.
        """

    @abstractmethod
    async def get_document_model_list(self) -> list:
        """
        Each factory should define what model it wants.
        """

    @abstractmethod
    @lru_cache()
    def get_app_config(self):
        """
        Each factory should define what config it wants.
        """

    def create_app(self, title=DEFAULT_APP_NAME) -> FastAPI:
        app = FastAPI(title=title, debug=self.get_app_config().DEBUG, version=self.API_VERSION)

        origins = ["*"]

        app.add_middleware(
            CORSMiddleware,
            allow_origins=origins,
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )

        self.__load_local_config()
        app.state.config = self.get_app_config()
        self.__setup_main_logger(app, level=logging.DEBUG)
        app.state.aws_session = aws.init_app(app)
        self.__setup_cloud_log(app)
        self.__load_cloud_config(app)

        mongodb = database.MongoDB(self.get_app_config().DATABASE_URL, self.get_app_config().DATABASE_NAME)

        @app.on_event("startup")
        async def startup_db_client():
            await mongodb.connect()
            document_model_list = await self.get_document_model_list()
            await init_beanie(database=app.state.db.get_database(), document_models=document_model_list)
            app.state.logger.info("Database connected")

        @app.on_event("shutdown")
        async def shutdown_db_client():
            await mongodb.close()
            app.state.logger.info("Database disconnected")

        app.state.db = mongodb
        app.state.config = self.get_app_config()
        self.__init_modules(app)
        self.__init_builtin_api(app)

        @app.exception_handler(RequestValidationError)
        async def validation_exception_handler(request, exc):
            print(exc)
            return await async_response(data=exc.detail, code="error_unprocessable_entity", message="Validation failed",
                                        status_code=exc.status_code)

        @app.exception_handler(Exception)
        async def http_exception_handler(request, exc):
            if isinstance(exc, InternalBaseException):
                detail = exc.detail
                return await async_response(data=detail.data, code=detail.code, message=detail.message,
                                            status_code=exc.status_code)
            else:
                app.state.logger.warn(f"Exception, request:{request.__dict__}, exc:{exc}")
                return await async_response(code="error_internal_server", message="Internal server error",
                                            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR)

        return app

    def __load_local_config(self):
        dotenv.load_dotenv(override=True)
        update_dict_with_cast(self.get_app_config(), os.environ)

    def __load_cloud_config(self, app):
        if not app.state.aws_session or not self.get_app_config().AWS_PARAMETER_PATH_PREFIX:
            app.state.logger.warn("No AWS session or Parameter Storage configuration, ignore cloud config")
            return

        cloud_conf = {}

        params = {
            "Path": self.get_app_config().AWS_PARAMETER_PATH_PREFIX,
            "Recursive": True,
            "WithDecryption": True
        }

        # AWS only give us 10 parameters per api call
        ssm_client = app.state.aws_session.client("ssm")
        while True:
            result = ssm_client.get_parameters_by_path(**params)
            cloud_conf.update({para["Name"].split("/")[-1]: para["Value"] for para in result["Parameters"]})
            if not result.get("NextToken"):
                break
            params.update({"NextToken": result["NextToken"]})

        update_dict_with_cast(self.get_app_config(), cloud_conf)

    def __init_modules(self, app):
        self.init_modules(app)

    def __setup_main_logger(self, app, logger_name=DEFAULT_LOGGER_NAME, level=logging.INFO):
        logger = self.__setup_logger(app, logger_name, level)
        app.state.logger = logger

    #
    def __setup_logger(self, app, logger_name, level=logging.INFO):
        logger = logging.getLogger(logger_name)
        logger.setLevel(level)

        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(logging.Formatter(fmt=LOG_FMT))
        logger.addHandler(stream_handler)

        return logger

    def __setup_cloud_log(self, app):
        if app.state.aws_session and self.get_app_config().AWS_LOGGROUP_NAME:
            logs_client = app.state.aws_session.client("logs")
            watchtower_handler = watchtower.CloudWatchLogHandler(
                log_group_name=self.get_app_config().AWS_LOGGROUP_NAME,
                boto3_client=logs_client, create_log_group=False)
            watchtower_handler.setFormatter(logging.Formatter(fmt=LOG_FMT_NO_DT, datefmt=LOG_DT_FMT))
            app.state.logger.addHandler(watchtower_handler)

    def __init_builtin_api(self, app):

        @app.get(f'/health', tags=["System"])
        async def health():
            return await async_response()

        @app.get(f'/hello', tags=["System"])
        async def hello():
            return await async_response(data={"API version": app.version})



# import base64
# import json
# from typing import Union
#
# from cryptography.hazmat.backends import default_backend
# from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
# from fastapi import HTTPException
# from fastapi.security import OAuth2PasswordBearer
# from jose import JWTError, jwt
#
# from jwt.const import JWT_AES_KEY, JWT_AES_IV
# from jwt.interface import JWTUserInterface, JWTTokenInterface
#
# oauth2_scheme_user = OAuth2PasswordBearer(tokenUrl=f"auth/user/cruisys_token")
#
#
# class JWTService:
#     @staticmethod
#     async def create_jwt_token(data: dict, app_config) -> str:
#         jwt_user_interface = JWTUserInterface(**data)
#         cipher_str = await JWTService.encrypt_aes_plain(jwt_user_interface)
#         jwt_token_interface = JWTTokenInterface(data=cipher_str)
#         token = jwt.encode(jwt_token_interface.to_dict(), app_config.JWT_SECRET_KEY,
#                            algorithm=app_config.JWT_ALGORITHM)
#         return token
#
#     @staticmethod
#     async def decode_jwt_token(token: str, app_config) -> Union[dict, JWTUserInterface, HTTPException]:
#         try:
#             payload = jwt.decode(token, app_config.JWT_SECRET_KEY, algorithms=[app_config.JWT_ALGORITHM])
#             plain_obj = await JWTService.decrypt_aes_cipher(JWTTokenInterface(**payload))
#             jwt_user_interface = JWTUserInterface(**plain_obj)
#             return jwt_user_interface
#         except JWTError:
#             raise HTTPException(status_code=401, detail="Invalid token")
#
#     @staticmethod
#     async def encrypt_aes_plain(interface: JWTUserInterface):
#         plaintext = json.dumps(interface.to_dict()).encode('utf-8')
#
#         cipher = Cipher(algorithms.AES(JWT_AES_KEY.encode('utf-8')), modes.CFB(JWT_AES_IV.encode('utf-8')),
#                         backend=default_backend())
#         encryptor = cipher.encryptor()
#         ciphertext = encryptor.update(plaintext) + encryptor.finalize()
#
#         return base64.b64encode(ciphertext).decode('utf-8')
#
#     @staticmethod
#     async def decrypt_aes_cipher(interface: JWTTokenInterface):
#         ciphertext = base64.b64decode(interface.data)
#
#         cipher = Cipher(algorithms.AES(JWT_AES_KEY.encode('utf-8')), modes.CFB(JWT_AES_IV.encode('utf-8')),
#                         backend=default_backend())
#         decryptor = cipher.decryptor()
#         plaintext = decryptor.update(ciphertext) + decryptor.finalize()
#
#         return json.loads(plaintext.decode('utf-8'))



from dataclasses import dataclass, asdict, fields


@dataclass
class InternalBaseInterface:
    def __init__(self, **kwargs):
        valid_fields = {f.name: kwargs.get(f.name) for f in fields(self)}
        for name, value in valid_fields.items():
            setattr(self, name, value)

    def to_dict(self):
        return asdict(self)



from pydantic_settings import BaseSettings


class BaseConfig(BaseSettings):
    DEBUG: bool = False
    RUN_PORT: int = 5000
    TIME_ZONE: str = "Asia/Taipei"

    # Request
    REQUEST_VERIFY_SSL: bool = True
    REQUEST_PROXY: str = ''
    REQUEST_RETRY: int = 5
    REQUEST_CONN_TIMEOUT: float = 5
    REQUEST_READ_TIMEOUT: float = 5
    REQUEST_WRITE_TIMEOUT: float = 5
    REQUEST_POOL_TIMEOUT: float = 5

    # AWS
    AWS_ACCESS_KEY_ID: str = ""
    AWS_SECRET_KEY: str = ""
    AWS_REGION: str = ""
    AWS_PARAMETER_PATH_PREFIX: str = ""
    AWS_LOGGROUP_NAME: str = ""

    # MongoDB
    DATABASE_URL: str = ""
    DATABASE_NAME: str = ""

    # Micro Service
    AUTH_BASE_URL: str = ""
    ORGANIZATION_BASE_URL: str = ""
    CUSTOMER_BASE_URL: str = ""
    CAR_BASE_URL: str = ""
    RELATIONSHIP_MANAGEMENT_BASE_URL: str = ""
    TICKET_BASE_URL: str = ""
    NOTIFY_BASE_URL: str = ""
    THIRD_PARTY_BASE_URL: str = ""
    SCHEDULER_BASE_URL: str = ""

    # JWT
    JWT_ALGORITHM: str = 'HS512'
    JWT_SECRET_KEY: str = ""
    JWT_ACCESS_TOKEN_EXPIRES_SECONDS: int = 86400
    JWT_REFRESH_TOKEN_EXPIRES_SECONDS: int = 604800  # 7days
    JWT_ERROR_MESSAGE_KEY: str = "message"

    class Config:
        case_sensitive = False



import json

from fastapi import status
from fastapi.responses import JSONResponse
from beanie import Document, Link


async def async_response(data=None, message=None, code=None, page_no=None, total_num=None, page_size=None,
                         status_code=status.HTTP_200_OK):
    def _serialize(data):
        # if issubclass(type(data), Document):
        if isinstance(data, Document):
            link_field_list = []
            for field_name in data.__annotations__:
                field_type = getattr(data, field_name)
                if isinstance(field_type, Link):
                    link_field_list.append(field_name)

            data = json.loads(data.model_dump_json())
            if link_field_list:
                for field_name in link_field_list:
                    if isinstance(data[field_name], dict) and "id" in data[field_name].keys():
                        data[field_name] = data[field_name]["id"]
        return data

    if isinstance(data, JSONResponse):
        return data

    ret = {}
    if isinstance(data, list):
        data = [_serialize(d) for d in data]
    else:
        data = _serialize(data)

    ret['code'] = code or "ok"

    ret['message'] = message or "success"

    if page_no and total_num and page_size:
        ret['data'] = {
            'page_no': page_no,
            'page_size': page_size,
            'total_num': total_num,
            'page_data': data
        }
    else:
        ret['data'] = data

    return JSONResponse(status_code=status_code, content=ret)



from dataclasses import dataclass

from ..interface.base_interface import InternalBaseInterface


@dataclass
class JWTUserInterface(InternalBaseInterface):
    id: str
    name: str
    phone: str
    mobile: str
    email: str
    title: str
    organization_parent: str
    organization_id: str
    ext_user_id: str
    ext_dealer_code: str
    ext_dept_code: str
    ext_dept_name: str
    line_id: str
    line_url: str
    role: str
    avatar: str = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


@dataclass
class JWTCustomerInterface(InternalBaseInterface):
    id: str
    ccid: str
    name: str
    phone: str
    mobile: str
    email: str
    title: str
    organization_parent: str
    organization_id: str
    ext_user_id: str
    ext_dealer_code: str
    ext_dept_code: str
    ext_dept_name: str
    line_id: str
    line_url: str
    role: str
    avatar: str = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


@dataclass
class JWTTokenInterface(InternalBaseInterface):
    data: str

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



from datetime import datetime
from enum import Enum
from typing import List, Tuple

import pymongo
from fastapi import FastAPI
from beanie import Document
from pydantic import Field

from ..const import DEF_PAGE_SIZE, DEF_PAGE_NO


class InternalBaseDocument(Document):
    create_time: datetime = Field(default_factory=datetime.utcnow)
    update_time: datetime = Field(default_factory=datetime.utcnow)

    @classmethod
    async def get_pagination_list(cls, app: FastAPI, query: list = None, sort: List[Tuple] = None,
                                  page_size: int = DEF_PAGE_SIZE, page_no: int = DEF_PAGE_NO, ignore_cache: bool = False,
                                  fetch_links: bool = False):
        if not query:
            final_query = []
        else:
            final_query = query

        if not sort:
            final_sort = [(cls.id, pymongo.ASCENDING)]
        else:
            final_sort = []
            for temp_sort in sort:
                if temp_sort[1] == OrderTypeEnum.ASC:
                    final_sort.append((temp_sort[0], pymongo.ASCENDING))
                elif temp_sort[1] == OrderTypeEnum.DESC:
                    final_sort.append((temp_sort[0], pymongo.DESCENDING))
                else:
                    print(f"order type value error: temp_sort:{temp_sort}")
                    continue

        total_num = await cls.find(*final_query, ignore_cache=ignore_cache, fetch_links=fetch_links).sort(
            *final_sort).count()
        if total_num == 0:
            page_data = []
        else:
            page_data = await cls.find(*final_query, ignore_cache=ignore_cache, fetch_links=fetch_links).sort(
                *final_sort).limit(page_size).skip((page_no - 1) * page_size).to_list()

        return page_no, page_size, total_num, page_data


class OrderTypeEnum(str, Enum):
    ASC = "ASC"
    DESC = "DESC"



import httpx

from fastapi import FastAPI

from ..exception.internal_exception import GatewayTimeoutException, BadGatewayException


async def async_request(app: FastAPI, method, url, **kwargs):
    timeout = httpx.Timeout(connect=app.state.config.REQUEST_CONN_TIMEOUT, read=app.state.config.REQUEST_READ_TIMEOUT,
                            write=app.state.config.REQUEST_WRITE_TIMEOUT, pool=app.state.config.REQUEST_POOL_TIMEOUT)

    try:
        async with httpx.AsyncClient(timeout=timeout) as client:
            response = await client.request(method, url, **kwargs)
            return response
    except httpx.TimeoutException as exc:
        app.state.logger.warn(
            f"async_request(), TimeoutException, exc: {exc}, url: {url}, method: {method}, kwargs: {kwargs}")
        raise GatewayTimeoutException(str(exc))
    except Exception as exc:
        app.state.logger.warn(
            f"async_request(), Exception, exc: {exc}, url: {url}, method: {method}, kwargs: {kwargs}")
        raise BadGatewayException(str(exc))



# -*- coding: UTF-8 -*-


DATE_FMT = "%Y-%m-%d"
TIME_FMT = "%H:%M"
DT_FMT = f"{DATE_FMT} {TIME_FMT}:%S"

DEFAULT_LOGGER_NAME = "uvicorn"
LOG_DT_FMT = f"{DT_FMT},%03d"
LOG_FMT_NO_DT = "[%(levelname)7s][%(name)s][%(filename)s::%(funcName)s(%(lineno)s)]: %(message)s"
LOG_FMT = "[%(asctime)s]" + LOG_FMT_NO_DT

ARR_EXPORT_SHORT_FMT = "MM/DD HH:mm"
ARR_EXPORT_DATE_FMT = "YYYY/MM/DD"
ARR_EXPORT_DATE_ONLY_FMT = "MM/DD"
ARR_DATE_FMT = "YYYY-MM-DD"
ARR_DATE_FMT_NO_DASH = "YYYYMMDD"
ARR_EXPORT_DATETIME_FMT = "YYYY/MM/DD HH:mm"
ARR_TIME_FMT = "HH:mm:ss"
ARR_TIME_ONLY_FMT = "HH:mm"
ARR_RFC3339_DT_FMT = f"{ARR_DATE_FMT}T{ARR_TIME_FMT}ZZ"
ARR_DATETIME_FMT_NO_DASH = "YYYYMMDDHHmmss"
ARR_HUMAN_READ_FMT = f"{ARR_DATE_FMT} {ARR_TIME_FMT}"
ARR_IMPORT_TS_FMT = "YYYYMMDDHHmmss.SSSSSS"
ARR_STR_DAY = "day"

STR_EMPTY = ""
STR_SPACE = " "
STR_DASH = "-"
STR_UNDERSCORE = "_"
STR_SLASH = "/"
STR_COMMA = ","

SECONDS_PER_YEAR = 30758400
SECONDS_PER_DAY = 86400
SECONDS_PER_MIN = 60

DEF_PAGE_NO = 1
DEF_PAGE_SIZE = 15
DEF_ORDER_BY = "_id"



from fastapi import status
from .base_exception import InternalBaseException


class BadGatewayException(InternalBaseException):
    code = "error_bad_gateway"
    message = "Bad gateway"

    def __init__(self, message: str = None, **kwargs):
        _message = message or self.message
        super().__init__(status.HTTP_502_BAD_GATEWAY, self.code, _message, **kwargs)


class GatewayTimeoutException(InternalBaseException):
    code = "error_gateway_timeout"
    message = "Gateway timeout"

    def __init__(self, message: str = None, **kwargs):
        _message = message or self.message
        super().__init__(status.HTTP_504_GATEWAY_TIMEOUT, self.code, _message, **kwargs)



import dictdiffer
from typing import Optional

from pydantic import BaseModel


class Operate(BaseModel):
    add: Optional[dict] = None
    remove: Optional[dict] = None
    change: Optional[dict] = None

    @classmethod
    async def generate_operate(cls, original: dict = None, compare: dict = None):
        if original:
            original = await cls.remove_ignore_field(original)
        else:
            original = {}

        if compare:
            compare = await cls.remove_ignore_field(compare)
        else:
            compare = {}

        diff_result = {'add': {}, 'remove': {}, 'change': {}}
        for diff in list(dictdiffer.diff(original, compare)):
            diff = (list(diff))
            if diff[0] == 'add' or diff[0] == 'remove':
                diff[2] = list(diff[2])
                for j in range(0, len(diff[2])):
                    diff[2][j] = list(diff[2][j])
            if diff[1] == 'change':
                diff[2] = list(diff[2])

            if diff[0] == 'change':
                diff_result['change'][diff[1]] = {'orig': diff[2][0], 'new': diff[2][1]}

            if diff[0] == 'add' or diff[0] == 'remove':
                for j in range(0, len(diff[2])):
                    diff_result[diff[0]][diff[2][j][0]] = diff[2][j][1]

        return Operate(**diff_result)

    @classmethod
    async def remove_ignore_field(cls, model_dict: dict):
        return {k: v for k, v in model_dict.items() if k not in ['create_time', 'update_time']}
